--
-- href.hs
--
-- Copyright (c) 2003 Minero Aoki <aamine@loveruby.net>
--
-- This program is free software.
-- You can distribute/modify this program under the terms of
-- the GNU LGPL, Lesser General Public License version 2.
-- For details of the GNU LGPL, see the file "COPYING".
--

module Main
  (main)
where

import System
import System.IO
import System.Directory
import Control.Monad
import Text.Regex
import Data.Char
import Data.List

import HrefUtils

hrefMaxColumns = 50 :: Int

main :: IO ()
main = do
    prefix <- myGetEnv hrefDataDirEnv
    args <- getArgs
    case args of
      []         -> do mods <- readModuleList prefix 
                       putList mods
      [name]     -> if isUpper (headChar name)
                       then lookupModule prefix name
                       else lookupFunction prefix "" name
      [mod,func] -> lookupFunction prefix mod func
      _          -> do hPutStrLn stderr "0..2 arguments must be given"
                       exitWith $ ExitFailure 1

lookupModule :: String -> String -> IO ()
lookupModule dbpath pattern = do
    mods <- completeModule dbpath pattern
    case mods of
      []    -> notFound "module" pattern
      [mod] -> do entries <- dirEntries (dbpath ++ "/" ++ encodePath mod)
                  putStrLn $ "---- module " ++ mod
                  putList $ map (decodePath) entries
      _     -> putList mods

completeModule :: String -> String -> IO [String]
completeModule dbpath pattern = do
    candidates <- readModuleList dbpath
    return $ filter (\c -> isPrefixOf (downcase pattern) (downcase c)) candidates

readModuleList :: String -> IO [String]
readModuleList dbpath = do
    cs <- readFile (dbpath ++ "/" ++ hrefKeyListFile)
    return $ sort.nub $ map (moduleName) $ lines cs

lookupFunction :: String -> String -> String -> IO ()
lookupFunction dbpath mod func = do
    candidates <- completeFunction dbpath mod func
    case candidates of
      []  -> notFound "function" $ mod ++ " " ++ func
      [f] -> describeFunction dbpath f
      _   -> case filter (\f -> strcasecmp func $ functionName f) candidates of
               [f] -> describeFunction dbpath f
               _   -> putList candidates

completeFunction :: String -> String -> String -> IO [String]
completeFunction dbpath mod func = do
    candidates <- readFunctionList dbpath
    return $ filter (matchp $ makeCompRegexp mod func) candidates

makeCompRegexp :: String -> String -> Regex
makeCompRegexp mod func =
    mkRegexWithOpts ("^" ++ mod ++ ".*," ++ func) True False

readFunctionList :: String -> IO [String]
readFunctionList dbpath = do
    cs <- readFile (dbpath ++ "/" ++ hrefKeyListFile)
    return $ lines cs

describeFunction :: String -> String -> IO ()
describeFunction dbpath func = do
    putStrLn func
    cs <- readFile $ dbpath ++"/"++
                     (encodePath $ moduleName func) ++"/"++
                     (encodePath $ functionName func)
    putStr cs

putList :: [String] -> IO ()
putList xs = do putList' hrefMaxColumns $ sort xs
                putStrLn ""

putList' :: Int -> [String] -> IO ()
putList' col [] = return ()
putList' col (x:xs)
    | col > 0   = do putStr (x ++ " ")
                     putList' (col - length x - 1) xs
    | otherwise = do putStrLn ""
                     putStr (x ++ " ")
                     putList' hrefMaxColumns xs

notFound :: String -> String -> IO ()
notFound kind name = do
    hPutStrLn stderr ("no such " ++ kind ++ ": " ++ name)
    exitWith $ ExitFailure 1

dirEntries :: FilePath -> IO [String]
dirEntries path = do
    entries <- getDirectoryContents path
    return $ filter (\e -> case e of
                             "." -> False
                             ".." -> False
                             _ -> True) entries

matchp :: Regex -> String -> Bool
matchp re str = case matchRegex re str of
                  Just _ -> True
                  Nothing -> False

strcasecmp :: String -> String -> Bool
strcasecmp a b = downcase a == downcase b

downcase :: String -> String
downcase str = map (toLower) str
